#!/usr/bin/env python3

import json
import string
import textwrap

import CifFile
import openpyxl

subscript = [
    '\N{Subscript Zero}',
    '\N{Subscript One}',
    '\N{Subscript Two}',
    '\N{Subscript Three}',
    '\N{Subscript Four}',
    '\N{Subscript Five}',
    '\N{Subscript Six}',
    '\N{Subscript Seven}',
    '\N{Subscript Eight}',
    '\N{Subscript Nine}',
]

bib = {}
bib['A3B_hP8_158_d_a'] = """\
@Article{A3B_hP8_158_d_a,
    author ="Fletcher, J. M. and Gardner, W. E. and Fox, A. C. and Topping, G.",
    title  ="X-Ray{,} infrared{,} and magnetic studies of α- and β-ruthenium trichloride",
    journal  ="J. Chem. Soc. A",
    year  ="1967",
    issue  ="0",
    pages  ="1038-1045",
    publisher  ="The Royal Society of Chemistry",
    doi  ="10.1039/J19670001038",
    url  ="http://dx.doi.org/10.1039/J19670001038"
}
"""  # noqa: E501

bib['A4B_oI20_74_beh_e'] = """\
@inbook{A4B_oI20_74_beh_e,
    place={Berlin &amp; Heidelberg, Germany},
    edition={8th},
    series={Gmelin Handbook of Inorganic and Organometallic Chemistry},
    title={Supplement Volume B2},
    volume={B2},
    booktitle={Alloys of Uranium with Alkali Metals, Alkaline Earths, and Elements of Main Groups III and IV},
    publisher={Springer},
    author={Borgstedt, Hans Ulrich and Wedemeyer, Horst and Buschbeck, Karl-Christian},
    year={1989},
    pages={1–333},
    collection={Gmelin Handbook of Inorganic and Organometallic Chemistry}
}

"""  # noqa: E501

bib['A4B_tP10_125_m_a'] = """\
@article{A4B_tP10_125_m_a,
    author = {R. Graham and G. C. S. Waghorn and P. T. Davies},
    title = {An X–ray investigation of the lead–platinum system},
    journal = {Acta Crystallographica},
    year = 1954,
    volume = {7},
    number = {10},
    pages = {634--635},
    month = {Oct},
    doi = {10.1107/S0365110X54002137},
    url = {https://doi.org/10.1107/S0365110X54002137},
}

"""

bib['A2B_oP12_62_2c_c.SrH2'] = """\
@inbook{A2B_oP12_62_2c_c.SrH2,
    author={R. C. Ropp},
    title={Encyclopedia of the Alkaline Earth Compounds},
    place={Oxford},
    publisher={Elsevier},
    year=2013,
    pages={30--31}
}

"""

bib['AB8C2_oC22_35_a_ab3e_e'] = """\
@article{AB8C2_oC22_35_a_ab3e_e,
    title = {CHEMICAL AND STRUCTURAL INVESTIGATION OF THE AB$sub 2$O$sub 8$
             COMPOUNDS OF Mo-V-O, U-V-O, AND U-Mo-O SYSTEMS.},
    author = {Mahe-Pailleret, P},
    abstractNote = {},
    doi = {},
    journal = {Rev. Chim. Miner. 7: 807-48(1970).},
    number = ,
    volume = ,
    place = {France},
    year = {1970},
    month = {1}
}

"""

bib['AB_hP6_144_a_a'] = """\
@article{doi:10.1063/1.46096,
    author = {Kusaba,Keiji  and Weidner,Donald J. },
    title = {Structure of high pressure phase I in ZnTe},
    journal = {AIP Conference Proceedings},
    volume = {309},
    number = {1},
    pages = {553-556},
    year = {1994},
    doi = {10.1063/1.46096},
    URL = {https://aip.scitation.org/doi/abs/10.1063/1.46096},
    eprint = {https://aip.scitation.org/doi/pdf/10.1063/1.46096}
}

"""

bib['A12B36CD12_cF488_210_h_3h_a_fg'] = """\
@article{A12B36CD12_cF488_210_h_3h_a_fg,
    author = {Tiritiris, Ioannis and Schleid, Thomas},
    title = {Synthesis, Crystal Structure, and Thermal Decomposition of Mg(H2O)6 [B12H12]×6H2O.},
    journal = {ChemInform},
    volume = {35},
    number = {25},
    pages = {},
    doi = {10.1002/chin.200425008},
    url = {https://onlinelibrary.wiley.com/doi/abs/10.1002/chin.200425008},
    eprint = {https://onlinelibrary.wiley.com/doi/pdf/10.1002/chin.200425008},
    abstract = {Abstract For Abstract see ChemInform Abstract in Full Text.},
    year = {2004}
}

"""  # noqa: E501

bib['AB3_tP32_86_g_3g'] = """\
@article{AB3_tP32_86_g_3g,
  title={PHASE DIAGRAM OF THE TITANIUM-PHOSPHORUS SYSTEM},
  author={EREMENKO, VN and LISTOVNICHII, VE},
  journal={DOPOVIDI AKAD NAUK UKR SSR},
  number={9},
  pages={1176--1179},
  year={1965}
}

"""

bib['AB32CD4E8_tP184_93_i_16p_af_2p_4p'] = """\
@article{AB32CD4E8_tP184_93_i_16p_af_2p_4p,
  title={Crystal and molecular structures of [AsPh 4][Ln (S 2 PMe 2) 4](Ln= Ce or Tm) and their comparison with results obtained from paramagnetic nuclear magnetic resonance data in solution},
  author={Spiliadis, Stavros and Pinkerton, A Alan and Schwarzenbach, Dieter},
  journal={Journal of the Chemical Society, Dalton Transactions},
  number={9},
  pages={1809--1813},
  year={1982},
  publisher={Royal Society of Chemistry}
}

"""  # noqa: E501

bib['A3B_oC64_66_gi2lm_2l'] = """\
@article{A3B_oC64_66_gi2lm_2l,
  title={Pressure-induced metallization of dense (H 2 S) 2 H 2 with high-Tc  superconductivity},
  author={Duan, Defang and Liu, Yunxian and Tian, Fubo and Li, Da and Huang, Xiaoli and Zhao, Zhonglong and Yu, Hongyu and Liu, Bingbing and Tian, Wenjing and Cui, Tian},
  journal={Scientific reports},
  volume={4},
  pages={6968},
  year={2014},
  publisher={Nature Publishing Group}
}

"""  # noqa: E501

bib['ABC2_hP4_164_a_b_d'] = """\
@phdthesis{ABC2_hP4_164_a_b_d,
  title={Intermetallic compounds by reductive annealing},
  author={Kift, Rebecca Louise},
  year={2010},
  school={University of Hull}
}

"""


def formula_subscripts(formula):
    result = ''
    for char in formula:
        if char in '0123456789':
            result += subscript[int(char)]
        else:
            result += char
    return result


def run(filename='prototypes.xlsx'):
    data = {}

    wb = openpyxl.load_workbook(filename=filename)
    ws = wb['Sheet1']

    row_max = ws.max_row

    data = {
        'prototype': [],
        'nSpecies': [],
        'nAtoms': [],
        'Pearson symbol': [],
        'Strukturbericht designation': [],
        'AFLOW prototype': [],
        'space group symbol': [],
        'space group number': [],
        'notes': []
    }
    hyperlink = []

    for row in ws.iter_rows(min_row=2, max_row=row_max):
        hyperlink.append(row[0].hyperlink.display)
        for cell, column in zip(row, data.keys()):
            data[column].append(cell.value)

    data['hyperlink'] = hyperlink

    # Now clean up the data
    tmp = []
    # Characters in notes
    for value in data['notes']:
        tmp.append(value.replace('\u00a0', ' '))
    data['notes'] = tmp

    # Subscripts in formulas of prototypes
    tmp = []
    for value in data['prototype']:
        value = formula_subscripts(value)
        print(value)
        tmp.append(value)
    data['prototype'] = tmp

    # doubling of Strukturbericht designations
    tmp = []
    for value in data['Strukturbericht designation']:
        if value is not None:
            length = int(len(value) / 2)
            value = value[0:length]
        tmp.append(value)
    data['Strukturbericht designation'] = tmp

    # doubling of space group symbols
    tmp = []
    for value in data['space group symbol']:
        if value is not None:
            length = int(len(value) / 2)
            value = value[0:length].replace('\u00af', '\N{Combining Overline}')
        print(value)
        tmp.append(value)
    data['space group symbol'] = tmp

    with open('prototypes.json', 'w') as fd:
        json.dump(data, fd, indent=4)


def download():
    """Download the CIF files from AFLOW"""
    # Get the data from the json
    with open('prototypes.json', 'r') as fd:
        data = json.load(fd)

    # Fetch the cif file
    for hyperlink in data['hyperlink']:
        url = urllib.parse.urlsplit(hyperlink)
        path = url.path
        root, extension = os.path.splitext(os.path.basename(path))
        filename = root + '.cif'
        new_path = os.path.join('prototypes', filename)
        if os.path.exists(new_path):
            continue
        new_url = urllib.parse.urljoin(hyperlink, 'CIF/' + filename)
        print(f'{new_url} --> {new_path}')
        try:
            with urllib.request.urlopen(new_url) as response:
                with open(new_path, 'w') as fd:
                    fd.write(response.read().decode("utf-8"))
        except urllib.error.HTTPError as e:
            print(f'#### Error: {e}')


def get_bibtex(prototype):
    """Create the BibTex for the prototype

    Example:

    @article{doi:10.1139/v68-576,
        author = {Knop, Osvald and Reid, K. I. G. and Sutarno and Nakagawa,
                 Yasuaki},
        title = {Chalkogenides of the transition elements. VI. X-Ray, neutron,
                 and magnetic investigation of the spinels Co3O4, NiCo2O4,
                 Co3S4, and NiCo2S4},
        journal = {Canadian Journal of Chemistry},
        volume = {46},
        number = {22},
        pages = {3463-3476},
        year = {1968},
    }
    """
    try:
        cif = CifFile.ReadCif(prototype + '.cif')
    except Exception:
        print(f'Error in {prototype}.cif\n\n')
        raise
    tmp = cif['findsym-output']
    authors = tmp['_publ_author_name']
    if '_journal_name_full' in tmp:
        journal = tmp['_journal_name_full']
    else:
        journal = tmp['_journal_name_full_name']
    if '_journal_volume' in tmp:
        volume = tmp['_journal_volume']
    else:
        volume = 'n/a'
        print(f'{prototype} has no journal volume')
    year = tmp['_journal_year']
    first = tmp['_journal_page_first']
    last = tmp['_journal_page_last']
    title = tmp['_publ_Section_title']

    author_list = ' and '.join(authors)
    pages = f'{first}-{last}'

    template = string.Template(
        """\
        @article{$prototype,
        author = {$author_list},
        title = {$title},
        journal = {$journal},
        volume = {$volume},
        pages = {$pages},
        year = $year,
        }
        """
    )

    citation = template.substitute(
        prototype=prototype,
        author_list=author_list,
        title=title.strip(),
        journal=journal.strip(),
        volume=volume,
        pages=pages,
        year=year
    )

    return citation


def test():
    import bibtexparser
    import reference_handler

    with open('references.bib') as fd:
        data = fd.read()

    tmp = bibtexparser.loads(data).entries_dict
    writer = bibtexparser.bwriter.BibTexWriter()
    biblio = {}
    for key, data in tmp.items():
        biblio[key] = writer._entry_to_bibtex(data)

    references = reference_handler.Reference_Handler(':memory:')
    for key, citation in biblio.items():
        references.cite(
            raw=citation,
            alias=key,
            module='parser',
            level=1,
            note=f"The principle citation for {key}."
        )

    citations = references.dump(fmt='text')
    for citation, text, count, level in citations:
        print(f"{citation} {textwrap.fill(text)}\n")


if __name__ == "__main__":  # pragma: no cover
    import os.path
    import urllib.error
    import urllib.parse
    import urllib.request

    test()

    # if True:
    #     test()
    # elif False:
    #     bibtex = ''
    #     with open('prototypes.json', 'r') as fd:
    #         data = json.load(fd)

    #     for prototype in data['AFLOW prototype']:
    #         print(prototype)
    #         if prototype in bib:
    #             text = bib[prototype]
    #         else:
    #             text = get_bibtex(prototype)
    #         bibtex += text
    #     with open('references2.bib', 'w') as fd:
    #         fd.write(bibtex)
    # elif False:
    #     run()
    # elif False:
    #     # Fix the prototypes to have added formula part in some cases
    #     # Get the data from the json
    #     with open('prototypes.json', 'r') as fd:
    #         data = json.load(fd)

    #     aflow_prototypes = []
    #     for hyperlink in data['hyperlink']:
    #         url = urllib.parse.urlsplit(hyperlink)
    #         path = url.path
    #         root, extension = os.path.splitext(os.path.basename(path))
    #         aflow_prototypes.append(root)

    #     data['AFLOW prototype'] = aflow_prototypes

    #     with open('prototypes2.json', 'w') as fd:
    #         json.dump(data, fd, indent=4)
    # else:
    #     # Get the data from the json
    #     with open('prototypes2.json', 'r') as fd:
    #         data = json.load(fd)

    #     cp = data['cell parameters'] = []
    #     sites = data['sites'] = []
    #     # Fetch the cif file to get the adjustable parameters
    #     for hyperlink in data['hyperlink']:
    #         url = urllib.parse.urlsplit(hyperlink)
    #         path = url.path
    #         root, extension = os.path.splitext(os.path.basename(path))
    #         filename = root + '.cif'
    #         new_path = os.path.join('.', filename)
    #         if not os.path.exists(new_path):
    #             print(f'{new_path} does not exist')
    #         else:
    #             try:
    #                 cif = CifFile.ReadCif(new_path)
    #             except Exception:
    #                 print(f'Error in {new_path}\n\n')
    #                 raise
    #             tmp = cif['findsym-output']
    #             params = tmp['_aflow_params'].split(',')
    #             values = tmp['_aflow_params_values'].split(',')

    #             cell_parameters = []
    #             for param, value in zip(params, values):
    #                 if param[0] in ('x', 'y', 'z'):
    #                     break
    #                 if param[0] == '\\':
    #                     cell_parameters.append((param[1:], value))
    #                 elif param == 'b/a':
    #                     cell_parameters.append(('b', tmp['_cell_length_b']))
    #                 elif param == 'c/a':
    #                     cell_parameters.append(('c', tmp['_cell_length_c']))
    #                 else:
    #                     cell_parameters.append((param, value))
    #             cp.append(cell_parameters)

    #             site_data = []
    #             for symbol, site, mult in zip(
    #                 tmp['_atom_site_type_symbol'],
    #                 tmp['_atom_site_Wyckoff_label'],
    #                 tmp['_atom_site_symmetry_multiplicity']
    #             ):
    #                 site_data.append((site, int(mult), symbol))
    #             sites.append(site_data)

    #     with open('prototypes3.json', 'w') as fd:
    #         json.dump(data, fd, indent=4)



# -*- coding: utf-8 -*-

"""Non-graphical part of the Crystal Builder step in a SEAMM flowchart
"""

try:
    import importlib.metadata as implib
except Exception:
    import importlib_metadata as implib
import logging
import pprint

import crystal_builder_step
import molsystem
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: 'job' and 'printer'. 'job' send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# 'printer' sends output to the file 'step.out' in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('Crystal Builder')


class CrystalBuilder(seamm.Node):
    """
    The non-graphical part of a Crystal Builder step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : CrystalBuilderParameters
        The control parameters for Crystal Builder.

    See Also
    --------
    TkCrystalBuilder,
    CrystalBuilder, CrystalBuilderParameters
    """

    def __init__(
        self, flowchart=None, title='Crystal Builder', extension=None
    ):
        """A step for Crystal Builder in a SEAMM flowchart.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
            flowchart: seamm.Flowchart
                The non-graphical flowchart that contains this step.

            title: str
                The name displayed in the flowchart.
            extension: None
                Not yet implemented
        """
        logger.debug('Creating Crystal Builder {}'.format(self))

        super().__init__(
            flowchart=flowchart,
            title='Crystal Builder',
            extension=extension,
            logger=logger
        )  # yapf: disable

        self.parameters = crystal_builder_step.CrystalBuilderParameters()

    @property
    def version(self):
        """The semantic version of this module.
        """
        return crystal_builder_step.__version__

    @property
    def git_revision(self):
        """The git version of this module.
        """
        return crystal_builder_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
            P: dict
                An optional dictionary of the current values of the control
                parameters.
        Returns
        -------
            description : str
                A description of the current step.
        """
        if not P:
            P = self.parameters.values_to_dict()

        aflow_prototype = P['AFLOW prototype']
        data = crystal_builder_step.prototype_data[aflow_prototype]

        cell = data['cell']
        sites = data['sites']
        is_common = False
        for prototype, aflow in crystal_builder_step.common_prototypes.items():
            if aflow == aflow_prototype:
                is_common = True
                break

        if is_common:
            text = "Building a {prototype} system"
        elif data['strukturbericht'] is not None:
            text = (
                'Building a system with the Struktubericht designation: '
                f"{data['strukturbericht']}"
            )
        else:
            text = (
                f'Building a system with the AFLOW prototype {aflow_prototype}'
            )

        text += '\n'

        text += f"\n            exemplar structure: {data['prototype']}"
        text += f"\n                         notes: {data['description']}"
        text += f"\n                Pearson symbol: {data['pearson_symbol']}"
        if data['strukturbericht'] is not None:
            text += (
                f"\n   Strukturbericht designation: {data['strukturbericht']}"
            )
        text += f"\n               number of atoms: {data['n_atoms']}"
        text += f"\n               AFLOW prototype: {aflow_prototype}"
        text += '\n'

        label = 'cell'
        for name, param0 in cell:
            value = P[name]
            text += f'\n                    {label:4} {name:>5}: {value}'
            label = ' '

        text += '\n'

        label = 'sites'
        for site_data, symbol in zip(sites, P['elements']):
            site, mult, _ = site_data
            site = f'{mult}{site}'
            text += f'\n                   {label:6} {site:>4}: {symbol}'
            label = ' '

        return self.header + '\n' + __(text, **P, indent=4 * ' ').__str__()

    def run(self):
        """Run a Crystal Builder step.

        Returns
        -------

        next_node : seamm.Node
            The next node object in the flowchart.

        """

        next_node = super().run(printer)

        # Get the system
        system_db = self.get_variable('_system_db')
        configuration = system_db.system.configuration

        # Get the values of the parameters, dereferencing any variables
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        self.logger.debug(f'Dereferenced values:\n{pprint.pformat(P)}')

        # Print what we are doing -- getting formatted values for printing
        PP = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data,
            formatted=True,
            units=False
        )
        self.logger.debug(f'Formatted values:\n{pprint.pformat(PP)}')
        printer.important(__(self.description_text(PP), indent=self.indent))

        # Create the configuration from the cif file for the prototype
        aflow_prototype = P['AFLOW prototype']

        package = self.__module__.split('.')[0]
        files = [p for p in implib.files(package) if aflow_prototype in str(p)]
        if len(files) > 0:
            path = files[0]
            data = path.read_text()
            configuration.from_cif_text(data)

        # Now set the cell parameters. If unmentioned the lattice parameters
        # get the previous value, e.g. a, a, c. The angles are those of the
        # prototype if not mentioned.
        a0, b0, c0, alpha0, beta0, gamma0 = configuration.cell.parameters

        data = crystal_builder_step.prototype_data[aflow_prototype]
        cell = data['cell']
        sites = data['sites']

        tmp = {
            'a': None,
            'b': None,
            'c': None,
            'alpha': alpha0,
            'beta': beta0,
            'gamma': gamma0
        }
        tmp0 = {**tmp}
        i = 0
        for name, param0 in cell:
            tmp0[name] = param0
            i += 1
            if name in ('a', 'b', 'c'):
                tmp[name] = P[name].to('Å').magnitude
            else:
                tmp[name] = P[name].to('degree')
            # Fill in any that were skipped ...
        last = None
        new_cell = []
        for name, param in tmp.items():
            if param is not None:
                last = param
            new_cell.append(last)

        self.logger.debug(f'cell = {new_cell}')
        configuration.cell.parameters = new_cell

        # And the elements for the sites. Not that symmetry has been lowered
        # to P1
        symbols = []
        for site_data, symbol in zip(sites, P['elements']):
            site, mult, symbol0 = site_data
            if self.is_expr(symbol):
                symbol = self.get_variable(symbol)
            symbols.extend([symbol] * mult)
        self.logger.debug(f'symbols = {symbols}')
        atnos = molsystem.elements.to_atnos(symbols)
        column = configuration.atoms.get_column('atno')
        column[0:] = atnos

        # Requested citations for the AFLOW protoype library
        self.references.cite(
            raw=self._bibliography['MEHL2017S1'],
            alias='MEHL2017S1',
            module='crystal_builder_step',
            level=1,
            note='Citation for the AFLOW library of prototype, part 1.'
        )
        self.references.cite(
            raw=self._bibliography['HICKS2019S1'],
            alias='HICKS2019S1',
            module='crystal_builder_step',
            level=1,
            note='Citation for the AFLOW library of prototype, part 2.'
        )

        # And citation for the structure itself:
        self.references.cite(
            raw=self._bibliography[aflow_prototype],
            alias=aflow_prototype,
            module='crystal_builder_step',
            level=1,
            note='Citation for the crystal strcuture of the prototype'
        )

        return next_node



# -*- coding: utf-8 -*-

"""
crystal_builder_step
A step for building crystals in SEAMM
"""

# Bring up the classes so that they appear to be directly in
# the crystal_builder_step package.

from crystal_builder_step.crystal_builder import CrystalBuilder  # noqa: F401, E501
from crystal_builder_step.crystal_builder_parameters import lattice_systems  # noqa: F401, E501
from crystal_builder_step.crystal_builder_parameters import CrystalBuilderParameters  # noqa: F401, E501
from crystal_builder_step.crystal_builder_step import CrystalBuilderStep  # noqa: F401, E501
from crystal_builder_step.tk_crystal_builder import TkCrystalBuilder  # noqa: F401, E501

from crystal_builder_step.crystal_metadata import common_prototypes  # noqa: F401, E501
from crystal_builder_step.crystal_metadata import prototypes  # noqa: F401
from crystal_builder_step.crystal_metadata import prototype_data  # noqa: F401

# Handle versioneer
from ._version import get_versions
__author__ = """Paul Saxe"""
__email__ = 'psaxe@molssi.org'
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



# -*- coding: utf-8 -*-

"""Non-graphical part of the Crystal Builder step in a SEAMM flowchart
"""

try:
    import importlib.metadata as implib
except Exception:
    import importlib_metadata as implib
import json

prototypes = {}
prototype_data = {}

common_prototypes = {
    'simple cubic (SC)': 'A_cP1_221_a',
    'body-centered cubic (BCC)': 'A_cI2_229_a',
    'face-centered cubic (FCC)': 'A_cF4_225_a',
    'diamond': 'A_cF8_227_a',
    'zincblende (ZnS)': 'AB_cF8_216_c_a',
    'wurtzite (ZnS)': 'AB_hP4_186_b_b'
}


def read_prototypes():
    """Read data for the AFLOW prototypes"""
    global prototypes
    global prototype_data

    # Read in the prototype metadata
    package = 'crystal-builder-step'
    files = [p for p in implib.files(package) if 'prototypes.json' in str(p)]
    if len(files) > 0:
        path = files[0]
        data = path.read_text()
        prototypes = json.loads(data)
    else:
        raise IOError('Prototypes JSON file not found!')

    prototypes['n_sites'] = []
    max_sites = 0
    # yapf: disable
    for (
        prototype,
        n_elements,
        n_atoms,
        pearson_symbol,
        strukturbericht,
        aflow,
        spacegroup,
        spacegroup_number,
        description,
        cell,
        sites
    ) in zip(
        prototypes['prototype'],
        prototypes['nSpecies'],
        prototypes['nAtoms'],
        prototypes['Pearson symbol'],
        prototypes['Strukturbericht designation'],
        prototypes['AFLOW prototype'],
        prototypes['space group symbol'],
        prototypes['space group number'],
        prototypes['notes'],
        prototypes['cell parameters'],
        prototypes['sites']
    ):
        prototypes['n_sites'].append(len(sites))
        if len(sites) > max_sites:
            max_sites = len(sites)
        prototype_data[aflow] = {
            'prototype': prototype,
            'n_elements': n_elements,
            'n_atoms': n_atoms,
            'pearson_symbol': pearson_symbol,
            'strukturbericht': strukturbericht,
            'aflow': aflow,
            'spacegroup': spacegroup,
            'spacegroup_number': spacegroup_number,
            'description': description,
            'cell': cell,
            'sites': sites
        }
    # yapf: enable
    prototypes['max_sites'] = max_sites


read_prototypes()



# -*- coding: utf-8 -*-

"""The graphical part of a Crystal Builder step"""

import tkinter as tk
import tkinter.ttk as ttk

import seamm
import seamm_widgets as sw
import crystal_builder_step  # noqa: F401


class TkCrystalBuilder(seamm.TkNode):
    """
    The graphical part of a Crystal Builder step in a flowchart.

    Attributes
    ----------
    namespace : str
        The namespace of the current step.
    node : Node
        The corresponding node of the non-graphical flowchart
    dialog : Dialog
        The Pmw dialog object
    sub_tk_flowchart : TkFlowchart
        A graphical Flowchart representing a subflowchart
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in Crystal Builder_parameters.py

    See Also
    --------
    CrystalBuilder, TkCrystalBuilder,
    CrystalBuilderParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.
        """
        self.dialog = None
        self._in_reset = False
        self._last_prototype = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h
        )

    def create_dialog(self):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the
        Crystal Builder_parameters module.

        See Also
        --------
        TkCrystalBuilder.reset_dialog
        """

        frame = super().create_dialog(title='Crystal Builder')

        # Shortcut for parameters
        P = self.node.parameters

        # Create the frames for cell and atom sites
        cell_frame = self['cell_frame'] = ttk.LabelFrame(
            self['frame'],
            borderwidth=4,
            relief='sunken',
            text='Cell',
            labelanchor='n',
            padding=10
        )
        self['site_frame'] = ttk.LabelFrame(
            self['frame'],
            borderwidth=4,
            relief='sunken',
            text='Atom Sites',
            labelanchor='n',
            padding=10
        )

        # The create the widgets
        for key in ('prototype_group', 'n_sites', 'prototype'):
            self[key] = P[key].widget(frame)
        for key in ('a', 'b', 'c', 'alpha', 'beta', 'gamma'):
            self[key] = P[key].widget(cell_frame)

        # And the sites for the current elements
        i = 0
        for symbol in P['elements'].get():
            i += 1
            key = f'site {i}'
            self[key] = sw.LabeledEntry(self['site_frame'], labeltext=key)
            self[key].set(symbol)

        # Setup bindings
        self['prototype_group'].combobox.bind(
            "<<ComboboxSelected>>", self.reset_dialog
        )
        self['prototype_group'].combobox.bind("<Return>", self.reset_dialog)
        self['prototype_group'].combobox.bind("<FocusOut>", self.reset_dialog)

        self['n_sites'].combobox.bind(
            "<<ComboboxSelected>>", self.reset_dialog
        )
        self['n_sites'].combobox.bind("<Return>", self.reset_dialog)
        self['n_sites'].combobox.bind("<FocusOut>", self.reset_dialog)

        self['prototype'].combobox.bind(
            "<<ComboboxSelected>>", self.reset_dialog
        )
        self['prototype'].combobox.bind("<Return>", self.reset_dialog)
        self['prototype'].combobox.bind("<FocusOut>", self.reset_dialog)

        # and lay them out
        self.reset_dialog()

    def reset_dialog(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Crystal Builder_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.

        Parameters
        ----------
        widget

        See Also
        --------
        TkCrystalBuilder.create_dialog
        """
        if self._in_reset:
            return
        self._in_reset = True

        # Remove any widgets previously packed
        frame = self['frame']
        for slave in frame.grid_slaves():
            slave.grid_forget()
        for slave in self['cell_frame'].grid_slaves():
            slave.grid_forget()
        for slave in self['site_frame'].grid_slaves():
            slave.grid_forget()

        prototype_group = self['prototype_group'].get()
        n_sites = self['n_sites'].get()
        prototype = self['prototype'].get()

        self._tmp = {}
        cb_prototypes = crystal_builder_step.prototypes
        if prototype_group == 'common':
            prototypes = [*crystal_builder_step.common_prototypes]
            self._tmp = {
                p: v for p, v in crystal_builder_step.common_prototypes.items()
            }
            self['prototype'].combobox.config(values=prototypes)
        elif prototype_group == 'Strukturbericht':
            prototypes = []
            for n_proto_sites, struk, proto, description, aflow in zip(
                    cb_prototypes['n_sites'],
                    cb_prototypes['Strukturbericht designation'],
                    cb_prototypes['prototype'],
                    cb_prototypes['notes'],
                    cb_prototypes['AFLOW prototype']
            ):  # yapf: disable
                if struk is not None:
                    if n_sites == 'any' or n_proto_sites == int(n_sites):
                        key = f'{struk}: {proto}: {description}'
                        prototypes.append(key)
                        self._tmp[key] = aflow
        else:
            prototypes = []
            for n_proto_sites, proto, description, aflow in zip(
                    cb_prototypes['n_sites'],
                    cb_prototypes['prototype'],
                    cb_prototypes['notes'],
                    cb_prototypes['AFLOW prototype']
            ):  # yapf: disable
                if n_sites == 'any' or n_proto_sites == int(n_sites):
                    key = f'{proto}: {description}'
                    prototypes.append(key)
                    self._tmp[key] = aflow
        prototypes.sort()
        self['prototype'].combobox.config(values=prototypes)
        if len(prototypes) == 0:
            width = 300
        else:
            width = max(len(x) for x in prototypes) + 5
        self['prototype'].config(width=width)
        if prototype in prototypes:
            self['prototype'].set(prototype)
        else:
            self['prototype'].combobox.current(0)
            prototype = self['prototype'].get()
        self._tmp['AFLOW prototype'] = self._tmp[prototype]

        # keep track of the row in a variable, so that the layout is flexible
        # if e.g. rows are skipped to control such as 'method' here
        row = 0
        widgets = []

        self['prototype_group'].grid(row=row, column=0, sticky=tk.EW)
        widgets.append(self['prototype_group'])
        row += 1

        if prototype_group != 'common':
            self['n_sites'].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self['n_sites'])
            row += 1

        self['prototype'].grid(row=row, column=0, sticky=tk.EW)
        widgets.append(self['prototype'])
        row += 1

        # Align the labels
        sw.align_labels(widgets)

        # And now the cell parameters
        self['cell_frame'].grid(row=row, column=0, sticky=tk.EW)
        row += 1

        aflow_prototype = self._tmp[prototype]
        cb_data = crystal_builder_step.prototype_data[aflow_prototype]
        cell_data = cb_data['cell']
        site_data = cb_data['sites']

        subrow = 0
        widgets = []
        for parameter, value in cell_data:
            w = self[parameter]
            w.grid(row=subrow, sticky=tk.EW)
            subrow += 1
            if aflow_prototype != self._last_prototype:
                w.set(value)
            widgets.append(w)
        sw.align_labels(widgets)

        # And the sites
        self['site_frame'].grid(row=row, column=0, sticky=tk.EW)
        row += 1

        subrow = 0
        widgets = []
        for site, mult, symbol in site_data:
            i = subrow + 1
            key = f'site {i}'
            if key not in self:
                self[key] = sw.LabeledEntry(self['site_frame'], labeltext=key)
            w = self[key]
            w.grid(row=subrow, sticky=tk.EW)
            subrow += 1
            if aflow_prototype != self._last_prototype:
                w.set(symbol)
            label = f'Site {i} -- {mult}{site}:'
            w.config(labeltext=label)
            widgets.append(w)
        sw.align_labels(widgets)

        # Remember the last prototype
        self._last_prototype = aflow_prototype

        self['frame'].grid_columnconfigure(0, weight=1, minsize=500)

        # All done resetting, so turn bindings back on.
        self._in_reset = False

    def right_click(self, event):
        """
        Handles the right click event on the node.

        See Also
        --------
        TkCrystalBuilder.edit
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def edit(self):
        """Present a dialog for editing the Crystal Builder input

        See Also
        --------
        TkCrystalBuilder.right_click
        """

        if self.dialog is None:
            P = self.node.parameters
            self._last_prototype = P['AFLOW prototype'].value
            self.create_dialog()

        self.dialog.activate(geometry='centerscreenfirst')

    def handle_dialog(self, result):
        """Handle the closing of the edit dialog

        What to do depends on the button used to close the dialog. If
        the user closes it by clicking the 'x' of the dialog window,
        None is returned, which we take as equivalent to cancel.

        Parameters
        ----------
        result : None or str
            The value of this variable depends on what the button
            the user clicked.

        """

        if result is None or result == 'Cancel':
            self.dialog.deactivate(result)
            return

        if result == 'Help':
            # display help!!!
            return

        if result != "OK":
            self.dialog.deactivate(result)
            raise RuntimeError(
                "Don't recognize dialog result '{}'".format(result)
            )

        self.dialog.deactivate(result)
        # Shortcut for parameters
        P = self.node.parameters

        # Get the values for all the widgets. This may be overkill, but
        # it is easy! You can sort out what it all means later, or
        # be a bit more selective.
        for key in P:
            if key != 'elements':
                P[key].set_from_widget()

        P['AFLOW prototype'].set(self._tmp['AFLOW prototype'])

        aflow_prototype = self._last_prototype
        cb_data = crystal_builder_step.prototype_data[aflow_prototype]
        site_data = cb_data['sites']
        i = 0
        elements = []
        for site, mult, symbol in site_data:
            i += 1
            key = f'site {i}'
            elements.append(self[key].get())
        P['elements'].set(elements)

        self._tmp = {}

    def handle_help(self):
        """Shows the help to the user when click on help button.

        """
        print('Help not implemented yet for Crystal Builder!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `crystal_builder_step` package."""

import pytest  # noqa: F401
import crystal_builder_step  # noqa: F401


def test_construction():
    """Just create an object and test its type."""
    result = crystal_builder_step.CrystalBuilder()
    assert str(type(result)) == (
        "<class 'crystal_builder_step.crystal_builder.CrystalBuilder'>"  # noqa: E501
    )



# -*- coding: utf-8 -*-
"""
Control parameters for the Crystal Builder step in a SEAMM flowchart
"""

import logging
import seamm

logger = logging.getLogger(__name__)


lattice_systems = {
    "cubic": {
        'lattices': ["fcc", "bcc", "primitive"],
        'parameters': ['a'],
        'cell': ['a', 'a', 'a', 90.0, 90.0, 90.0]
    },
    "hexagonal": {
        'lattices': ["primitive", "hcp"],
        'parameters': ['a', 'c'],
        'cell': ['a', 'a', 'c', 90.0, 90.0, 120.0]
    },
    "rhombohedral": {
        'lattices': ["primitive"],
        'parameters': ['a', 'alpha'],
        'cell': ['a', 'a', 'a', 'alpha', 'alpha', 'alpha']
    },
    "tetragonal": {
        'lattices': ["primitive", "body-centered"],
        'parameters': ['a', 'c'],
        'cell': ['a', 'a', 'c', 90.0, 90.0, 90.0]
    },
    "orthorombic": {
        'lattices': [
            "primitive",
            "body-centered",
            "base-centered",
            "face-centered",
        ],
        'parameters': ['a', 'b', 'c'],
        'cell': ['a', 'b', 'c', 90.0, 90.0, 90.0]
    },
    "monoclinic": {
        'lattices': ["primitive", "base-centered"],
        'parameters': ['a', 'b', 'c', 'beta'],
        'cell': ['a', 'b', 'c', 90.0, 'beta', 90.0]
    },
    "triclinic": {
        'lattices': ["primitive"],
        'parameters': ['a', 'b', 'c', 'alpha', 'beta', 'gamma'],
        'cell': ['a', 'b', 'c', 'alpha', 'beta', 'gamma']
    },
}


class CrystalBuilderParameters(seamm.Parameters):
    """
    The control parameters for Crystal Builder.

    The developer will add a dictionary of Parameters to this class.
    The keys are parameters for the current plugin, which themselves
    might be dictionaries.

    You need to replace the 'time' example below with one or more
    definitions of the control parameters for your plugin and application.

    Attributes
    ----------
    parameters : {'kind', 'default', 'default_units', 'enumeration',
                  'format_string', description', help_text'}
        A dictionary containing the parameters for the current step.
        Each key of the dictionary is a dictionary that contains the
        the following keys: kind, default, default_units, enumeration,
        format_string, description and help text.

    parameters['kind']: custom
        Specifies the kind of a variable. While the 'kind' of a variable might
        be a numeric value, it may still have enumerated custom values
        meaningful to the user. For instance, if the parameter is
        a convergence criterion for an optimizer, custom values like 'normal',
        'precise', etc, might be adequate. In addition, any
        parameter can be set to a variable of expression, indicated by having
        '$' as the first character in the field. For example, $OPTIMIZER_CONV.

    parameters['default'] : 'integer' or 'float' or 'string' or 'boolean' or
        'enum' The default value of the parameter, used to reset it.

    parameters['default_units'] : str
        The default units, used for resetting the value.

    parameters['enumeration']: tuple
        A tuple of enumerated values.

    parameters['format_string']: str
        A format string for 'pretty' output.

    parameters['description']: str
        A short string used as a prompt in the GUI.

    parameters['help_text']: tuple
        A longer string to display as help for the user.

    See Also
    --------
    CrystalBuilder, TkCrystalBuilder, CrystalBuilder
    CrystalBuilderParameters, Crystal BuilderStep

    Examples
    --------
    parameters = {
        "time": {
            "default": 100.0,
            "kind": "float",
            "default_units": "ps",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Simulation time:",
            "help_text": ("The time to simulate in the dynamics run.")
        },
    }
    """

    parameters = {
        "prototype_group": {
            "default": 'common',
            "kind": "enumeration",
            "default_units": None,
            "enumeration": ('common', 'Strukturbericht', 'all'),
            "format_string": "",
            "description": "Which prototypes:",
            "help_text": ("The group of prototypes to display")
        },
        "n_sites": {
            "default": 'any',
            "kind": "int",
            "default_units": None,
            "enumeration": ('any',),
            "format_string": "",
            "description": "Number of occupied sites:",
            "help_text": ("The number of occupied sites in the structure.")
        },
        "prototype": {
            "default": "",
            "kind": "enumeration",
            "default_units": None,
            "enumeration": ('place', 'holders'),
            "format_string": "",
            "description": "Prototype:",
            "help_text": ("The prototype to use")
        },
        "AFLOW prototype": {
            "default": "",
            "kind": "string",
            "default_units": None,
            "enumeration": tuple(),
            "format_string": "",
            "description": "AFLOW Prototype:",
            "help_text": ("The AFLOW prototype")
        },
        "elements": {
            "default": ['Al'],
            "kind": "list",
            "default_units": None,
            "enumeration": tuple(),
            "format_string": "",
            "description": "Elements:",
            "help_text": ("The elements at the different sites.")
        },
        "a": {
            "default": 10.0,
            "kind": "float",
            "default_units": "Å",
            "enumeration": None,
            "format_string": ".2f",
            "description": "a:",
            "help_text": ("The length of the first side of the cell.")
        },
        "b": {
            "default": 10.0,
            "kind": "float",
            "default_units": "Å",
            "enumeration": None,
            "format_string": ".2f",
            "description": "b:",
            "help_text": ("The length of the second side of the cell.")
        },
        "c": {
            "default": 10.0,
            "kind": "float",
            "default_units": "Å",
            "enumeration": None,
            "format_string": ".2f",
            "description": "c:",
            "help_text": ("The length of the third side of the cell.")
        },
        "alpha": {
            "default": 90.0,
            "kind": "float",
            "default_units": "degree",
            "enumeration": None,
            "format_string": ".1f",
            "description": "alpha:",
            "help_text": ("The angle between a and b.")
        },
        "beta": {
            "default": 90.0,
            "kind": "float",
            "default_units": "degree",
            "enumeration": None,
            "format_string": ".1f",
            "description": "beta:",
            "help_text": ("The angle between a and c.")
        },
        "gamma": {
            "default": 90.0,
            "kind": "float",
            "default_units": "degree",
            "enumeration": None,
            "format_string": ".1f",
            "description": "gamma:",
            "help_text": ("The angle between b and c.")
        },
    }

    def __init__(self, defaults={}, data=None):
        """
        Initialize the parameters, by default with the parameters defined above

        Parameters
        ----------
            defaults: dict
                A dictionary of parameters to initialize. The parameters
                above are used first and any given will override/add to them.
            data: dict
                A dictionary of keys and a subdictionary with value and units
                for updating the current, default values.

        """

        logger.debug('CrystalBuilderParameters.__init__')

        super().__init__(
            defaults={**CrystalBuilderParameters.parameters, **defaults},
            data=data
        )
